'''
Created on 2020-07-05

@author: wf
'''
import matplotlib.pyplot as plt
from collections import Counter
import numpy as np
import os

class Plot(object):
    '''
    create Plot based on counters
    see https://stackoverflow.com/questions/19198920/using-counter-in-python-to-build-histogram
    '''
    def __init__(self, valueList,title,xlabel=None,ylabel=None,gformat='.png',fontsize=12,plotdir=None,debug=False):
        '''
        Constructor
        '''
        self.counter=Counter(valueList)
        self.valueList=valueList
        self.title=title
        self.xlabel=xlabel
        self.ylabel=ylabel
        self.fontsize=fontsize
        self.gformat=gformat
        self.debug=debug
        path=os.path.dirname(__file__)
        if plotdir is not None:
            self.plotdir=plotdir
        else:
            self.plotdir=path+"/../plots/"
            os.makedirs(self.plotdir,exist_ok=True)
            
    def titleMe(self):   
        ''' set my title and labels '''     
        plt.title(self.title, fontsize=self.fontsize)
        if self.xlabel is not None:
            plt.xlabel(self.xlabel)
        if self.ylabel is not None:    
            plt.ylabel(self.ylabel)
            
    def showMe(self,mode='show',close=True):
        ''' show me in the given mode '''
        if mode=="show":
            plt.show() 
        else:
            plt.savefig(self.plotdir+self.title+self.gformat)
        if close:    
            plt.close()    
            
    def barchart(self,mode='show'):
        ''' barchart based histogram for the given counter '''
        labels, values = zip(*self.counter.items())
        indexes = np.arange(len(labels))
        width = 1
        self.titleMe()
        plt.bar(indexes, values, width)
        plt.xticks(indexes + width * 0.5, labels)
        plt.yticks(np.arange(1,max(values)+1,step=1))
        self.showMe(mode)
        
    def showDebug(self):    
        print("   value  list: ",self.valueList)
        print("counter  items: ",self.counter.items())
        print("counter values: ",self.counter.values())
        print("counter   keys: ",self.counter.keys())
        
    def hist(self,mode="show"):
        ''' create histogram for the given counter '''
        if self.debug:
            self.showDebug()
        self.titleMe()
        # see https://stackoverflow.com/a/2162045/1497139
        plt.hist(self.valueList,bins=len(self.counter.keys()))
        self.showMe(mode)
        pass
            
        


'''
Created on 2021-01-26

@author: wf
'''
from collections import Counter

class SchemaManager(object,):
    '''
    a manager for schemas
    '''
    def __init__(self,schemaDefs=None,baseUrl:str=None):
        '''
        constructor
            Args:
                schemaDefs(dict): a dictionary of schema names
                baseUrl(str): the base url to use for links
        '''
        self.baseUrl=baseUrl
        self.schemasByName={}
        if schemaDefs is None:
            schemaDefs={}
        for key,name in schemaDefs.items():
            self.schemasByName[key]=Schema(key,name) 
        pass

class Schema(object):
    '''
    a relational Schema
    '''

    def __init__(self,name:str,title:str):
        '''
        Constructor
        
        Args:
            name(str): the name of the schema
            title(str): the title of the schema
        '''
        self.name=name
        self.title=title
        self.propsByName={}
        
    @staticmethod    
    def generalizeColumn(tableList,colName:str):
        ''' 
        remove the column with the given name from all tables in the tablelist and
        return it
        
        Args:
            tableList(list): a list of Tables
            colName(string): the name of the column to generalize
        
        Returns:
            string: the column having been generalized and removed
        '''
        gCol=None
        for table in tableList:
            for col in table['columns']: 
                if col['name']==colName:
                    gCol=col.copy()
                    # no linking yet @FIXME - will need this later
                    if 'link' in gCol:
                        gCol.pop('link')
                    # is generalization protected for this column?
                    if not 'special' in col or not col['special']:
                        table['columns'].remove(col)
        return gCol                
        
    @staticmethod    
    def getGeneral(tableList,name:str,debug:bool=False):
        '''
        derive a general table from the given table list
        Args:
            tableList(list): a list of tables
            name(str): name of the general table
            debug(bool): True if column names should be shown
            
        Returns:
            at table dict for the generalized table
        '''
        general={'name':name,'columns':[]}
        colCount=Counter()
        for table in tableList:
            for col in table['columns']:
                columnId="%s.%s" % (col['name'],col['type'])
                if debug:
                    print (columnId) 
                colCount[columnId]+=1
        for columnId,count in colCount.items():
            if count==len(tableList): 
                colName=columnId.split('.')[0]
                generalCol=Schema.generalizeColumn(tableList, colName)
                general['columns'].append(generalCol)    
        return general
    
    @staticmethod
    def getGeneralViewDDL(tableList,name:str,debug=False)->str:
        ''' 
        get the DDL statement to create a general view
        
        Args:
            tableList: the list of tables
            name(str): the name of the view
            debug(bool): True if debug should be set
        '''
        general=Schema.getGeneral(tableList, name, debug)
        cols=""
        delim=""
        for col in general['columns']:
            cols+="%s%s" % (delim,col['name'])
            delim=","
        ddl="CREATE VIEW %s AS \n" % name
        delim=""
        for table in tableList:
            ddl+="%s  SELECT %s FROM %s" % (delim,cols,table['name'])
            delim="\nUNION\n"
        return ddl
        


'''
Created on 2020-08-22

@author: wf
'''
import os
import yaml
#from wikibot.mwTable import MediaWikiTable
# redundant copy in this library to avoid dependency issues
# original is at 
from lodstorage.mwTable import MediaWikiTable
class Query(object):
    ''' a Query e.g. for SPAQRL '''
    
    def __init__(self,name,query,lang='sparql',debug=False):
        '''
        constructor 
        Args:
            name(string): the name of the query
            query(string): the native Query text e.g. in SPARQL
            lang(string): the language of the query e.g. SPARQL
            debug(boolean): true if debug mode should be switched on
        '''
        self.name=name
        self.lang=lang
        self.query=query
        self.debug=debug
        
    def asWikiSourceMarkup(self):
        '''
        convert me to Mediawiki markup for syntax highlighting using the "source" tag
        
        Returns:
            string: the Markup
        '''
        markup="<source lang='%s'>\n%s\n</source>\n" %(self.lang,self.query)
        return markup
        
    def asWikiMarkup(self,listOfDicts):
        '''
        convert the given listOfDicts result to MediaWiki markup
        Args:
            listOfDicts(list): the list of Dicts to convert to MediaWiki markup
        Returns:
            string: the markup
        '''
        if self.debug:
            print(listOfDicts)
        mwTable=MediaWikiTable()
        mwTable.fromListOfDicts(listOfDicts)
        markup=mwTable.asWikiMarkup()        
        return markup

class QueryManager(object):
    '''
    manages prepackaged Queries
    '''

    def __init__(self,lang='sql',debug=False,path=None):
        '''
        Constructor
        Args:
            lang(string): the language to use for the queries sql or sparql
            debug(boolean): True if debug information should be shown
        '''
        self.queriesByName={}
        self.lang=lang
        self.debug=debug
        queries=QueryManager.getQueries(path=path)
        for name,queryDict in queries.items():
            if self.lang in queryDict:
                queryText=queryDict[self.lang]
                query=Query(name,queryText,lang=self.lang,debug=self.debug)
                self.queriesByName[name]=query
    
    @staticmethod
    def getQueries(path=None):
        if path is None:
            path=os.path.dirname(__file__)+"/.."
        queriesPath=path+"/queries.yaml"
        with open(queriesPath, 'r') as stream:
            examples = yaml.safe_load(stream)
        return examples
        


'''
Created on 2021-06-13

@author: wf
'''
from tabulate import tabulate

class TabulateCounter(object):
    '''
    helper for tabulating Counters
    '''


    def __init__(self, counter):
        '''
        Constructor
        '''
        self.counter=counter
    
    def mostCommonTable(self,headers=["#","key","count","%"],tablefmt='pretty',limit=50):
        '''
        get the most common Table
        '''
        bins=len(self.counter.keys())
        limit=min(bins,limit)
        total=sum(self.counter.values())
        binTable=[("total",bins,total)]
        for i,bintuple in enumerate(self.counter.most_common(limit)):
            key,count=bintuple
            binTable.append((i+1,key,count,count/total*100.0))
        
        table=tabulate(binTable,headers=headers,tablefmt=tablefmt,floatfmt=".2f")
        return table


'''
Created on 2020-08-24

@author: wf
'''
import urllib.request
import json
from lodstorage.jsonable import JSONAble, JSONAbleList
from datetime import date, datetime

class Sample(object):
    '''
    Sample dataset generator
    '''
    cityList=None

    def __init__(self):
        '''
        Constructor
        '''
        
    @staticmethod         
    def getSample(size):         
        listOfDicts=[]
        for index in range(size):
            listOfDicts.append({'pkey': "index%d" %index, 'cindex': index})
        return listOfDicts    
    
    @staticmethod
    def getCountries():
        countryJsonUrl="https://gist.githubusercontent.com/erdem/8c7d26765831d0f9a8c62f02782ae00d/raw/248037cd701af0a4957cce340dabb0fd04e38f4c/countries.json"
        with urllib.request.urlopen(countryJsonUrl) as url:
            countryList=json.loads(url.read().decode())
        return countryList
    
    @staticmethod
    def getCities():
        '''
        get a list of cities
        '''
        if Sample.cityList is None:
            cityJsonUrl="https://raw.githubusercontent.com/lutangar/cities.json/master/cities.json"
            with urllib.request.urlopen(cityJsonUrl) as url:
                Sample.cityList=json.loads(url.read().decode())
            for city in Sample.cityList:
                city['cityId']="%s-%s" % (city['country'],city['name'])    
        return Sample.cityList
    
    @staticmethod
    def dob(isoDateString):
        ''' get the date of birth from the given iso date state'''
        #if sys.version_info >= (3, 7):
        #    dt=datetime.fromisoformat(isoDateString)
        #else:
        dt=datetime.strptime(isoDateString,"%Y-%m-%d")  
        return dt.date()   
    
    @staticmethod
    def getRoyals():
        return Royal.getSamples()
    
    @staticmethod
    def getRoyalsInstances():
        lod=Royal.getSamples()
        royals=[]
        for record in lod:
            royal=Royal()
            royal.fromDict(record)
            royals.append(royal)
        return royals
    
class Royals(JSONAbleList):
    '''
    a non ORM Royals list
    '''
    def __init__(self,load=False):
        super(Royals, self).__init__("royals",clazz=None)
        if load:
            self.royals=Royal.getSamples()
        else:
            self.royals=None
        
class RoyalsORMList(JSONAbleList):
    def __init__(self,load=False):
        super(RoyalsORMList, self).__init__("royals",Royal)
        if load:
            self.royals=Sample.getRoyalsInstances()
            
class Royal(JSONAble):
    '''
    i am a single Royal
    '''
    
    @classmethod
    def getSamples(cls):
        listOfDicts=[
            {'name': 'Elizabeth Alexandra Mary Windsor', 'born': Sample.dob('1926-04-21'), 'numberInLine': 0, 'wikidataurl': 'https://www.wikidata.org/wiki/Q9682' },
            {'name': 'Charles, Prince of Wales',         'born': Sample.dob('1948-11-14'), 'numberInLine': 1, 'wikidataurl': 'https://www.wikidata.org/wiki/Q43274' },
            {'name': 'George of Cambridge',              'born': Sample.dob('2013-07-22'), 'numberInLine': 3, 'wikidataurl': 'https://www.wikidata.org/wiki/Q1359041'},
            {'name': 'Harry Duke of Sussex',             'born': Sample.dob('1984-09-15'), 'numberInLine': 6, 'wikidataurl': 'https://www.wikidata.org/wiki/Q152316'}
        ]
        today=date.today()
        for person in listOfDicts:
            born=person['born']
            age=(today - born).days / 365.2425
            person['age']=age
            person['ofAge']=age>=18
            person['lastmodified']=datetime.now()
        return listOfDicts
    
    def __repr__(self):
        text=self.__class__.__name__
        attrs=["name","born"]
        delim=":"
        for attr in attrs:
            if hasattr(self, attr):
                value=getattr(self,attr)
                text+=f"{delim}{value}"
                delim=":" 
        return text
         
    
class Cities(JSONAbleList):
    def __init__(self,load=False):
        super(Cities, self).__init__("cities",clazz=None)
        if load:
            self.cities=Sample.getCities()      
        else:
            self.cities=None  